import logging

from shapely.geometry import LinearRing
from teklia_toolbox.pagexml import PageXmlPage

from arkindex_worker.models import Element
from arkindex_worker.worker.metadata import MetaType

logger = logging.getLogger(__name__)

MAX_POLYGON_POINTS = 512


class PageXmlParser:
    def __init__(self, worker, path_or_xml, image):
        self.pagexml_page = PageXmlPage(path_or_xml)
        self.image = image
        self.worker = worker

    def simplify_polygon(self, region):
        if region.points is None:
            logger.warning(f"No points in region {region.id}")
            return

        # Ignore polygons that are completely outside the image
        if not any(
            0 < x <= self.image["width"] and 0 < y <= self.image["height"]
            for x, y in region.points
        ):
            logger.warning(
                f"Ignoring region {region.id} (this polygon is entirely outside the image's bounds)"
            )
            return

        # Crop polygons that are partially outside the image
        cropped_polygon = [
            [
                min(self.image["width"], max(0, x)),
                min(self.image["height"], max(0, y)),
            ]
            for x, y in region.points
        ]

        if len(cropped_polygon) < 3:
            logger.warning(
                f"Ignoring region {region.id} (not enough points in polygon)"
            )
            return

        polygon = LinearRing(cropped_polygon)

        for tolerance in range(1, 20):
            simplified = polygon.simplify(tolerance)
            points = list(simplified.coords)
            if len(points) <= MAX_POLYGON_POINTS:
                # The methods from base worker expect a list of lists, not tuples
                points = [list(point) for point in points]
                return points

        logger.warning(
            f"Ignoring region {region.id} (no points simplification possible using normal tolerances)"
        )

    def check_transcription(self, region):
        if region.text is None or not region.text.strip():
            logger.warning(f"No text in region {region.id}")
            return False

        return True

    def save(self, parent_element, image_id, page_type, paragraph_type, line_type):
        # The metadata can be missing if `exportTranscriptMetadata` is set to `false` in an export
        if (
            self.pagexml_page.metadata is None
            or self.pagexml_page.metadata.transkribus_metadata is None
        ):
            logger.error("Missing TranskribusMetadata XML tag in transcript")
            return None, []

        element = self.worker.create_sub_element(
            Element(parent_element),
            type=page_type,
            name=str(self.pagexml_page.metadata.transkribus_metadata.page_number),
            image=image_id,
            slim_output=False,
        )
        metadata = [
            {
                "name": "transkribus_page_id",
                "type": MetaType.Reference,
                "value": str(self.pagexml_page.metadata.transkribus_metadata.page_id),
            },
            {
                "name": "folio",
                "type": MetaType.Text,
                "value": str(
                    self.pagexml_page.metadata.transkribus_metadata.page_number
                ),
            },
            {
                "name": "transkribus_status",
                "type": MetaType.Text,
                "value": str(self.pagexml_page.metadata.transkribus_metadata.status),
            },
        ]
        self.worker.create_metadata_bulk(Element(element), metadata)

        if self.pagexml_page.page.text_regions is None or not len(
            self.pagexml_page.page.text_regions
        ):
            logger.warning("No transcriptions to save")
            return element, []

        region_count, line_count, region_ts_count, line_ts_count = 0, 0, 0, 0

        # Create paragraphs from text regions in a single call
        region_payloads = [
            (
                region,
                {"type": paragraph_type, "name": str(i), "polygon": polygon},
            )
            for i, region in enumerate(self.pagexml_page.page.text_regions)
            if (polygon := self.simplify_polygon(region))
        ]

        if not region_payloads:
            logger.warning("No valid regions to save")
            return element, []

        logger.debug("Creating region elements")
        regions_response = self.worker.create_elements(
            Element(element), [payload for _, payload in region_payloads]
        )
        region_count += len(regions_response)

        transcription_payloads = []

        # Add transcription payloads for regions with non-empty text
        for (region, _), region_element in zip(
            region_payloads, regions_response, strict=False
        ):
            if self.check_transcription(region):
                # Entities can be on several lines, so the reading order is important to merge entities correctly.
                # It is sent in the transcriptions.json file for the entity build task,
                # so we store it along the payload and we'll link it with the transcription ID after creating.
                tag_reading_order = next(
                    filter(lambda tag: tag.name == "readingOrder", region.tags)
                )
                index_region = int(tag_reading_order.index)

                transcription_payloads.append(
                    (
                        {
                            "element_id": region_element["id"],
                            "text": region.text,
                            "confidence": 1.0
                            if region.confidence is None
                            else region.confidence,
                        },
                        index_region,
                    )
                )
                region_ts_count += 1

            # Create line elements in a single call
            line_payloads = [
                (line, {"polygon": polygon, "type": line_type, "name": str(i)})
                for i, line in enumerate(region.lines)
                if (polygon := self.simplify_polygon(line))
            ]
            if line_payloads:
                logger.debug(f"Creating line elements in region {region.id}")
                # Cannot use base worker's create_elements because we do not have the full region elements, only ID,
                # and create_elements requires a zone.
                lines_response = self.worker.api_client.request(
                    "CreateElements",
                    id=region_element["id"],
                    body={
                        "worker_run_id": self.worker.worker_run_id,
                        "elements": [payload for _, payload in line_payloads],
                    },
                )
                line_count += len(lines_response)

                for index, (line, _) in enumerate(line_payloads):
                    transkribus_id = line.get_text("@id")
                    if transkribus_id is not None:
                        self.worker.create_metadata(
                            Element(lines_response[index]),
                            type=MetaType.Reference,
                            name="transkribus_id",
                            value=transkribus_id,
                        )

                # Add transcription payloads for lines with non-empty text
                for (line, _), line_element in zip(
                    line_payloads, lines_response, strict=False
                ):
                    if self.check_transcription(line):
                        transcription_payloads.append(
                            (
                                {
                                    "element_id": line_element["id"],
                                    "text": line.text,
                                    "confidence": 1.0
                                    if line.confidence is None
                                    else line.confidence,
                                },
                                # No reading order here, as lines are ignored for entity building
                                None,
                            )
                        )
                        line_ts_count += 1

        # Create all transcriptions for the page's regions and lines at once
        transcriptions = []
        if transcription_payloads:
            logger.debug("Creating transcriptions")
            response = self.worker.create_transcriptions(
                [payload for (payload, _) in transcription_payloads]
            )
            # Build the transcriptions.json payload
            transcriptions = [
                {"index_region": index_region, "transcription_id": transcription["id"]}
                for (_, index_region), transcription in zip(
                    transcription_payloads, response, strict=False
                )
                if index_region is not None
            ]

        logger.info(
            f"Created {region_count} {paragraph_type} elements with {region_ts_count} transcriptions "
            f"and {line_count} {line_type} elements with {line_ts_count} transcriptions."
        )
        return element, transcriptions
